////////////////////////////////////////////////////////////////////// 
// lamp/JointModel.cpp 
// (c) 2000-2006 Goncalo Abecasis
// 
// This file is distributed as part of the LAMP source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile LAMP.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Saturday July 15, 2006
// 
 
#include "ModelSummary.h"
#include "MapFunction.h"
#include "MathGenMin.h"
#include "StringHash.h"
#include "JointModel.h"
#include "MathStats.h"
#include "TreeInfo.h"
#include "Houdini.h"
#include "Conquer.h"
#include "Random.h"

#include <math.h>

#define   MODEL_NULL          0
#define   MODEL_EQUILIBRIUM   1
#define   MODEL_GENERAL       2

#define   RESCALE_THRESHOLD   (pow(2.0,-32))
#define   RESCALE_AMOUNT      (pow(2.0, 32))

bool   JointModel::quiet = true;

double JointModel::PREVALENCE = 0.05;
bool   JointModel::ADDITIVE = false;
bool   JointModel::DOMINANT = false;
bool   JointModel::RECESSIVE = false;
bool   JointModel::MULTIPLICATIVE = false;
bool   JointModel::FREE = true;

int    JointModel::ROUNDS = 3;
bool   JointModel::NELDERMEAD = true;
bool   JointModel::FLETCHERREEVES = false;
bool   JointModel::STOCHASTIC = false;
double JointModel::PRECISION = 1e-8;

bool   JointModel::UNRELATEDS = false;
bool   JointModel::SIBPAIRS = false;
bool   JointModel::AUTOMATIC = true;
bool   JointModel::NOCONSTRAINTS = false;

int    JointModel::MAXBITS = 16;
bool   JointModel::BUFFERING = true;
bool   JointModel::SKIPCAUSAL = false;

JointModel::JointModel()
   {
   families = 0;
   mantra = NULL;
   left = right = singlepoint = NULL;
   buffer = NULL;
   mendelFlag = false;
   }

JointModel::~JointModel()
   {
   FreeMemory();
   }

void JointModel::FreeMemory()
   {
   for (int i = 0; i < families; i++)
      if (singlepoint != NULL)
         {
         delete [] left[i];
         delete [] right[i];
         delete [] singlepoint[i];
         }

   if (mantra != NULL) delete [] mantra;
   if (left != NULL) delete [] left;
   if (right != NULL) delete [] right;
   if (singlepoint != NULL) delete [] singlepoint;
   if (buffer != NULL) delete [] buffer;

   mantra = NULL;
   left = right = singlepoint = NULL;
   }

bool JointModel::CheckMaps(MarkerMap & frame, MarkerMap & snps)
   {
   if (snps.markers.Length() <= 0)
      {
      printf("The candidate map is empty\n");
      return false;
      }

   if (frame.markers.Length() && frame.chromosome != snps.chromosome)
      {
      error("Candidate SNPs and markers in framework are in different chromosomes\n");
      return false;
      }

   bool success = true;

   StringHash hash;

   for (int i = 0; i < frame.markers.Length(); i++)
      hash.Add(frame.markers[i]);

   for (int i = 0; i < snps.markers.Length(); i++)
      if (hash.Find(snps.markers[i]) != -1)
         printf("Marker %s is a candidate and also in the framework map\n",
               (const char *) snps.markers[i], success = false);

   for (int i = 0; i < frame.markers.Length(); i++)
      if (Pedigree::LookupMarker(frame.markers[i]) == -1)
         printf("Framework marker %s is not in the pedigree\n",
                (const char *) frame.markers[i], success = false);

//   If a "candidate snp" is not in the pedigree, we simply evaluate
//   the evidence for linkage at its location.
//   for (int i = 0; i < snps.markers.Length(); i++)
//      if (Pedigree::LookupMarker(snps.markers[i]) == -1)
//         printf("Candidate marker %s is not in the pedigree\n",
//                (const char *) snps.markers[i], success = false);

   return success;
   }

void JointModel::AllocateMemory(Pedigree & ped, MarkerMap & frame)
   {
   int n = families = ped.familyCount;
   int m = frame.markers.Length();

   if (n == 0)
      return;

   mantra = new Mantra [n];

   if (m)
      {
      singlepoint = new Tree * [n];
      right = new Tree * [n];
      left = new Tree * [n];
      }

   for (int i = 0; i < n; i++)
      {
      mantra[i].Prepare(ped, *ped.families[i]);

      if (m == 0) continue;

      left[i] = new Tree [m];
      right[i] = new Tree [m];
      singlepoint[i] = new Tree [m];
      }

   buffer = new Tree [n];

   BuildBitsFilter(ped);
   }

void JointModel::ScoreSinglepoint(Pedigree & ped, MarkerMap & frame)
   {
   int n = ped.familyCount;
   int m = frame.markers.Length();

   IntArray id;
   id.Dimension(m);

   for (int i = 0; i < m; i++)
      id[i] = ped.LookupMarker(frame.markers[i]);

   printf("Scoring markers in framework map ... ");
   fflush(stdout);

   FuzzyInheritanceTree houdini;

   for (int i = 0; i < n; i++)
      {
      for (int j = 0; j < m; j++)
         {
         mantra[i].SelectMarker(id[j]);
         houdini.FuzzyScoreVectors(mantra[i]);

         if (houdini.WeightedSum() == 0.0)
            {
            singlepoint[i][j].MakeMinimalTree(1.0, mantra[i].bit_count);
            mendelFlag = true;
            }
         else
            singlepoint[i][j].Copy(houdini);
         }
      }

   printf(" Done!\n");
   }

void JointModel::ScoreConditionals(Pedigree & ped, MarkerMap & frame)
   {
   int n = ped.familyCount;
   int m = frame.markers.Length();

   if (m == 0) return;

   // Setup array with recombination fractions
   Vector theta;
   theta.Dimension(frame.positions.Length() - 1);

   for (int i = 0; i < frame.positions.Length() - 1; i++)
      theta[i] = DistanceToRecombination(frame.positions[i + 1] - frame.positions[i]);

   // Setup array to keep track of rescalings, which are used
   // to avoid underflows
   leftScaling.Dimension(m);
   rightScaling.Dimension(m);

   leftScaling.Zero();
   rightScaling.Zero();

   // Proceed to score left and right conditional probabilities ...
   printf("Scoring left and right conditional probabilities ... ");
   fflush(stdout);

   Multipoint engine;
   TreeInfo   stats;

   for (int i = 0; i < n; i++)
      if (!bitsFilter[i])
         {
         // printf("Family %s   \r", (const char *) ped.families[i]->famid);

         right[i][m - 1].CarbonCopy(singlepoint[i][m - 1]);

         for (int j = m - 1; j > 0; j--)
            {
            engine.CarbonCopy(right[i][j]);
            stats.GetTreeInfo(engine);
            engine.MoveAlong(mantra[i], theta[j - 1],
                             stats.mean < RESCALE_THRESHOLD ? RESCALE_AMOUNT : 1.0);
            engine.Multiply(singlepoint[i][j - 1]);

            right[i][j - 1].Copy(engine);

            if (stats.mean < RESCALE_THRESHOLD)
               rightScaling[j - 1]++;

            // printf("Likelihood %g\n", right[i][j - 1].WeightedSum());
            }

         if (right[i][0].WeightedSum() == 0.0)
            error("The likelihood for family %s is zero\n",
                  (const char *) ped.families[i]->famid);

         left[i][0].CarbonCopy(singlepoint[i][0]);

         for (int j = 0; j < m - 1; j++)
            {
            engine.CarbonCopy(left[i][j]);
            stats.GetTreeInfo(engine);

            engine.MoveAlong(mantra[i], theta[j],
                             stats.mean < RESCALE_THRESHOLD ? RESCALE_AMOUNT : 1.0);
            engine.Multiply(singlepoint[i][j + 1]);

            left[i][j + 1].Copy(engine);

            if (stats.mean < RESCALE_THRESHOLD)
               leftScaling[j + 1]++;

            // printf("Likelihood %g\n", left[i][j - 1].WeightedSum());
            }
         }

   printf(" Done!\n");
   }

GeneralMinimizer * JointModel::GetMinimizer()
   {
   if (NELDERMEAD)
      return new AmoebaMinimizer;

   if (FLETCHERREEVES)
      return new FletcherMinimizer;

   return new EvolutionaryMinimizer;
   }

void JointModel::OpenDetailFiles()
   {
   details[0].OpenFile("lamp-base.out");
   details[0].BaseModelHeader();
   details[1].OpenFile("lamp-linkage.out");
   details[1].LinkageHeader();
   details[2].OpenFile("lamp-indirect-association.out");
   details[2].IndirectAssociationHeader();
   details[3].OpenFile("lamp-direct-association.out");
   details[3].DirectAssociationHeader();
   }

void JointModel::CloseDetailFiles()
   {
   for (int i = 0; i < 4; i++)
      details[i].CloseFile();
   }

void JointModel::PrintDetails(int model, String & marker, String & trait, double llk, int parameters)
   {
   model = min(model, 3);

   details[model].loglikelihood = llk;
   details[model].fittedParameters = parameters;
   details[model].markerId = currentMarker;
   details[model].trait = trait;
   details[model].marker = marker;
   details[model].penetrances[0] = lk.penetrance[0];
   details[model].penetrances[1] = lk.penetrance[1];
   details[model].penetrances[2] = lk.penetrance[2];
   details[model].frequencies = lk.frequencies;
   details[model].OutputModel(currentModel);
   }

void JointModel::Analyze(Pedigree & ped, MarkerMap & frame, MarkerMap & candidates)
   {
   if (ped.affectionCount == 0) return;

   qtlMode = false;

   OptimizerInterface optim;
   Vector             point, transformedPoint;

   GeneralMinimizer * minimizer = GetMinimizer();

   optim.SetObjectiveFunction(*this);

   // IBD probabilities based on framework marker data
   linkage = new Tree [ped.familyCount];

   OpenDetailFiles();

   if (!UNRELATEDS)
      printf("\n"
             "                                TEST FOR         TEST FOR      TEST FOR OTHER \n"
             "                                LINKAGE         ASSOCIATION   LINKED VARIANTS \n"
             "                            ---------------- ---------------- ----------------\n"
             "LOCATION      TRAIT            LOD df pvalue    LOD df pvalue    LOD df pvalue\n"
             "==============================================================================\n");
   else
      printf("\n"
             "                                        TEST FOR    \n"
             "                                       ASSOCIATION  \n"
             "                                    ----------------\n"
             "LOCATION      TRAIT         ALLELE     LOD df pvalue   \n"
             "====================================================\n");

   double linkagePeak = -1.0, linkageDf;
   double associationPeak = -1.0, associationDf;
   String linkageMarker, linkageTrait, associationMarker, associationTrait;

   // Cycle through each candidate marker
   for (int i = 0; i < candidates.markers.Length(); i++)
      {
      currentMarker = ped.LookupMarker(candidates.markers[i]);
      markerInfo = currentMarker >= 0 ? ped.GetMarkerInfo(currentMarker) : NULL;

      MendelFilter(ped);
      ScoreLinkage(ped, frame, candidates.positions[i]);

      // We evaluate three models at each location
      //    model 0 -- no effect, maximize only allele frequencies
      //    model 1 -- allow linkage, but not linkage disequilibrium
      //    model 2 -- most general model
      //    model 3 and above -- assume that a specific allele directly affects
      //                         susceptibility or is in complete LD with an
      //                         allele that does
      // const  char * models[4] = {"Base", "Link", "Assoc", "Causal"};

      // Cycle through each affection status ...
      for (int j = 0; j < Pedigree::affectionCount; j++)
         {
         int modelCount = markerInfo == NULL ? 2 : (3 + markerInfo->CountAlleles());

         currentAffection = j;

         if (modelCount == 5 && !DOMINANT && !RECESSIVE)
            modelCount = 4;

         if (modelCount > 3 && SKIPCAUSAL)
            modelCount = 3;

         if (UNRELATEDS && modelCount == 2)
            continue;

         printf("%-13.13s %-13.13s ", (const char *) candidates.markers[i],
                                      (const char *) Pedigree::affectionNames[j]);
         fflush(stdout);

         for (int model = 0; model < modelCount; model++)
            {
            if (UNRELATEDS && model == 1) continue;
            if (UNRELATEDS && model == 2) continue;

            lk.SelectAffection(j);
            lk.SelectMarker(currentMarker);

            SelectModel(model);

            GetStartingPoint(point);

            SetupConstraints(optim, point);

            minimizer->Reset(optim.CountFreeParameters(), 1.0);
            minimizer->func = &optim;

            optim.BackTranslate(point, minimizer->point);
            minimizer->Minimize(PRECISION);
            optim.Translate(minimizer->point, point);

            double fmin = minimizer->fmin;
            Vector bestPoint = point;

            for (int round = 1; round < ROUNDS; round++)
               {
               GetStartingPoint(point, round);
               SetupConstraints(optim, point);

//               double fstart = Evaluate(point);

               minimizer->Reset(optim.CountFreeParameters(), 1.0);
               optim.BackTranslate(point, minimizer->point);
               minimizer->Minimize(PRECISION);

               if (minimizer->fmin < fmin)
                  {
                  fmin = minimizer->fmin;
                  optim.Translate(minimizer->point, bestPoint);
                  }
               }

            point = bestPoint;

            SetModelParameters(point);

            fmin -= adjustment;
            PrintDetails(currentModel,
                         candidates.markers[i], Pedigree::affectionNames[j],
                         -fmin, optim.CountFreeParameters());

            if (currentModel == 0)
               continue;
            else if (currentModel == 1)
               {
               double lr = details[1].loglikelihood - details[0].loglikelihood;
               int    df = details[1].fittedParameters - details[0].fittedParameters;

               PrintLOD(lr, df);

               if (lr > linkagePeak)
                  {
                  linkagePeak = lr;
                  linkageDf = df;
                  linkageMarker = candidates.markers[i];
                  linkageTrait = ped.affectionNames[j];
                  }
               }
            else if (currentModel == 2)
               {
               double lr = details[2].loglikelihood - details[1].loglikelihood;
               int    df = details[2].fittedParameters - details[1].fittedParameters;

               PrintLOD(lr, df);

               if (lr > associationPeak)
                  {
                  associationPeak = lr;
                  associationDf = df;
                  associationMarker = candidates.markers[i];
                  associationTrait = ped.affectionNames[j];
                  }
               }
            else if (UNRELATEDS)
               {
               if (currentAllele > 0)
                  printf("\n%28s", "");
               printf("%6s  ", (const char *) markerInfo->GetAlleleLabel(currentAllele + 1));

               double lr = details[currentModel].loglikelihood - details[0].loglikelihood;
               int    df = details[currentModel].fittedParameters - details[0].fittedParameters;

               if (lr > associationPeak)
                  {
                  associationPeak = lr;
                  associationDf = df;
                  associationMarker = candidates.markers[i];
                  associationTrait = ped.affectionNames[j];
                  }

               PrintLOD(lr, df);
               }
            else
               {
               if (modelCount > 4)
                  printf("\n  -- BASELINE: direct association with allele %-16.16s",
                          (const char *) markerInfo->GetAlleleLabel(currentAllele + 1));
               PrintLOD(details[2].loglikelihood - details[3].loglikelihood,
                        details[2].fittedParameters - details[3].fittedParameters);
               }
            }

         printf("\n");
         }
      }

   printf("\n");

   if (mendelFlag)
      printf("Mendelian inconsistencies encountered in some families\n"
             "   During, analysis genotypes that were not Mendelian consistent\n"
             "   were treated as missing\n\n");

   if (associationPeak > -1.0)
      printf("Peak association was LOD = %.3f (%.0f df) at %s for %s\n",
              associationPeak / log(10.0), associationDf,
              (const char *) associationMarker,
              (const char *) associationTrait);

   if (linkagePeak > -1.0)
      printf("Peak linkage was LOD = %.3f (%.0f df) at %s for %s\n",
              linkagePeak / log(10.0), linkageDf,
              (const char *) linkageMarker,
              (const char *) linkageTrait);

   printf("\n");
   CloseDetailFiles();

   delete [] linkage;
   delete minimizer;
   }

void JointModel::AnalyzeQTL(Pedigree & ped, MarkerMap & frame, MarkerMap & candidates)
   {
   if (ped.traitCount == 0) return;

   qtlMode = true;

   OptimizerInterface optim;
   Vector             point, transformedPoint;

   GeneralMinimizer * minimizer = GetMinimizer();

   optim.SetObjectiveFunction(*this);

   // IBD probabilities based on framework marker data
   linkage = new Tree [ped.familyCount];

   printf("\n"
          "                                                Trait Means                  \n"
          "                                         ------------------------            \n"
          "   Trait   Locus   Model   LogLk  p(-)   G=+/+    G=+/-    G=-/-   Al p(-|Al)\n"
          "============================================================================\n");

   // Cycle through each candidate marker
   for (int i = 0; i < candidates.markers.Length(); i++)
      {
      currentMarker = ped.LookupMarker(candidates.markers[i]);
      markerInfo = currentMarker >= 0 ? ped.GetMarkerInfo(currentMarker) : NULL;

      MendelFilter(ped);
      ScoreLinkage(ped, frame, candidates.positions[i]);

      // We evaluate three models at each location
      //    model 0 -- no effect, maximize only allele frequencies
      //    model 1 -- allow linkage, but not linkage disequilibrium
      //    model 2 -- most general model
      const char * models[4] = {"Base", "Link", "Assoc", "Causal"};
      double lkBase = 0.0;

      int modelCount = markerInfo == NULL ? 2 : (3 + markerInfo->CountAlleles());
      // Cycle through each quantitative trait ...
      for (int j = 0; j < Pedigree::traitCount; j++)
         {
         for (int model = 0; model < modelCount; model++)
            {
            SelectModel(model);

            qtl.SelectTrait(j);
            qtl.SelectMarker(currentMarker);

            GetStartingPoint(point);

            SetupConstraints(optim, point);

            minimizer->Reset(optim.CountFreeParameters());
            minimizer->func = &optim;

            optim.BackTranslate(point, minimizer->point);

            if (STOCHASTIC)
               {
               EvolutionaryMinimizer * miner = (EvolutionaryMinimizer *) minimizer;

               for (int i = 0; i < miner->points.rows / 2; i++)
                  {
                  GetStartingPoint(point, i);
                  SetupConstraints(optim, point);
                  optim.BackTranslate(point, miner->points[i]);
                  }

               miner->generate_random_points = false;
               }

            minimizer->Minimize(PRECISION);
            optim.Translate(minimizer->point, point);

            double fmin = minimizer->fmin;
            Vector bestPoint = minimizer->point;

            for (int round = 1; round < ROUNDS; round++)
               {
               GetStartingPoint(point, round);
               SetupConstraints(optim, point);

               // double fstart = Evaluate(point);

               minimizer->Reset(optim.CountFreeParameters(), 1.0);
               optim.BackTranslate(point, minimizer->point);
               minimizer->Minimize(PRECISION);

               if (minimizer->fmin < fmin)
                  {
                  fmin = minimizer->fmin;
                  bestPoint = minimizer->point;
                  }

               // printf("   %.4f -> %.4f\n", fstart, minimizer.fmin);
               }

            optim.Translate(bestPoint, point);

            if (currentModel == 0)
               lkBase = fmin;

            if (currentModel == 0)
               {
               // printf("%8.8s%8s%10.6f [REFERENCE MODEL]\n",
               //    (const char *) candidates.markers[i], models[currentModel],
               //    lkBase);
               continue;
               }

            printf("%8.8s%8.8s%8s%8.4f",
                   (const char *) Pedigree::traitNames[j],
                   (const char *) candidates.markers[i], models[currentModel],
                   lkBase - fmin);

            SetModelParameters(point);
            printf("%6.3f %8.3f %8.3f %8.3f",
                   disease[1], qtl.means[0], qtl.means[1], qtl.means[2]);

            if (currentModel < 2)
               {
               printf("\n");
               continue;
               }

            printf("%4s %5.3f\n",
                   (const char *) markerInfo->GetAlleleLabel(1),
                   qtl.frequencies[1] / (qtl.frequencies[0] + qtl.frequencies[1]));

            for (int i = 1; i < markerInfo->CountAlleles(); i++)
            printf("%69s %5.3f\n",
                   (const char *) markerInfo->GetAlleleLabel(i + 1),
                   qtl.frequencies[i * 2 + 1] / (qtl.frequencies[i * 2] + qtl.frequencies[i * 2 + 1]));
            }
         }
      }

   if (mendelFlag)
      printf("Mendelian inconsistencies encountered in some families\n"
             "   During, analysis genotypes that were not Mendelian consistent\n"
             "   were treated as missing\n\n");

   delete minimizer;
   delete [] linkage;
   }


void JointModel::GetStartingPoint(Vector & point, int round)
   {
   int freqs = markerInfo == NULL ? 1 : markerInfo->CountAlleles() * 2 - 1;

   point.Dimension(2 + freqs);

   double penetrances1[] = {0.20, 0.95, 0.05};
   double penetrances2[] = {0.20, 0.50, 0.50};

   double means1[] = {0.0, -0.90, 0.90};
   double means2[] = {0.0, 0.0, 0.0};

   if (round < 3)
      {
      point[0] = qtlMode ? means1[round] : penetrances1[round];
      point[1] = qtlMode ? means2[round] : penetrances2[round];
      }
   else
      {
      point[0] = globalRandom.Uniform(0.02, 0.98);
      point[1] = globalRandom.Uniform(0.02, 0.98);
      }

   // Setup the allele frequencies
   if (markerInfo != NULL)
      {
      // Calculate the sum of all allele frequencies
      double sum = 0.0;

      for (int j = 0; j < freqs / 2 + 1; j++)
         sum += markerInfo->freq[j + 1] + 1e-6;
      sum = 1.0 / sum;

      // Scale each frequency according to the sum of previous frequencies
      double cumulative = 0.0;

      for (int j = 0; j < freqs / 2; j++)
         {
         double freq = (markerInfo->freq[j + 1] + 1e-6) * sum;

         point[2 + j] = freq / (1.0 - cumulative);
         cumulative  += freq;
         }
      }

   // Allele frequencies for disease allele conditional on marker
   // allele frequencies
   for (int j = 0; j <= freqs / 2; j++)
      if (currentModel == 1 && SIBPAIRS && FREE)
         point[2 + freqs / 2 + j] = sqrt(PREVALENCE);
      else
         point[2 + freqs / 2 + j] =
             currentModel != 3 ? ((round < 3) ? sqrt(PREVALENCE) : globalRandom.Uniform(0.10,0.90)):
            (currentAllele == j) ? 1.0 : 0.0;
   }

void JointModel::SetupConstraints(OptimizerInterface & optim, Vector & point)
   {
   int markerAlleles = currentMarker >= 0 ? markerInfo->freq.Length() - 1 : 1;
   int freqs         = markerAlleles * 2 - 1;

   optim.Dimension(2 + freqs);

   // First we setup constraints for the two parameters used to model
   // penetrances ...
   if (currentModel == 0)
      {
      // Penetrances are fixed for null model
      optim.Fix(0, point[0]);
      optim.Fix(1, point[1]);
      }
   else
      {
      // TODO -- Fix penetrances sensibly when model is constrained (e.g. to set increasing allele)
      // Otherwise, fitted using one or two parameters, depending on genetic model
      optim.SetRange(0, qtlMode ? -1.0 : 0.0, 1.0);

      if (FREE)
         optim.SetRange(1, qtlMode ? 1.0 : 0.0, 1.0);
      else
         optim.Fix(1, point[1]);
      }

   // Marker allele frequencies are always free to vary
   for (int i = 0; i < markerAlleles - 1; i++)
      optim.SetRange(2 + i, 0.0, 1.0);

   // For calculating disease allele frequencies, we either
   //     estimate no parameters                   (model 0)
   //     estimate one parameter                   (model 1)
   //     estimate one parameter per marker allele (model 2)
   //     fix these at 1.0 or 0.0                  (model 3)
   for (int i = 0; i < markerAlleles; i++)
      if (currentModel == 3)
         optim.Fix(1 + markerAlleles + i, i == currentAllele ? 1.0 : 0.0);
      else if (currentModel == 0 || currentModel == 1 && i != markerAlleles - 1 ||
               currentModel == 1 && SIBPAIRS && FREE )
         optim.Fix(1 + markerAlleles + i, point[1 + markerAlleles + i]);
      else
         optim.SetRange(1 + markerAlleles + i, 0.000001, 0.999999);
   }

void JointModel::ScoreLinkage(Pedigree & ped, MarkerMap & frame, double position)
   {
   Multipoint engine;

   adjustment = 0;

   int flank = 0;

   // First find the flanking markers
   while (flank < frame.markers.Length() && position > frame.positions[flank])
      flank++;
   flank--;

   // Next score IBD probabilities using framework marker data
   if (flank >= 0)
      {
      // Male and female recombination fractions
      double theta[2];

      theta[0] = theta[1] = DistanceToRecombination(position - frame.positions[flank]);

      for (int j = 0; j < ped.familyCount; j++)
         if (!bitsFilter[j]) 
            {
            engine.CarbonCopy(left[j][flank]);
            engine.MoveAlong(mantra[j], theta, 1.0);

            linkage[j].Copy(engine);
            }

      adjustment += -leftScaling.Sum(0, flank) * log(RESCALE_AMOUNT);
      }
   else
      for (int j = 0; j < ped.familyCount; j++)
         linkage[j].MakeMinimalTree(1.0, mantra[j].bit_count);

   if (flank < frame.markers.Length() - 1)
      {
      double theta[2];

      theta[0] = theta[1] = DistanceToRecombination(frame.positions[flank+1] - position);

      for (int j = 0; j < ped.familyCount; j++)
         if (!bitsFilter[j]) 
            {
            engine.CarbonCopy(right[j][flank + 1]);
            engine.MoveAlong(mantra[j], theta, 1.0);

            linkage[j].Multiply(engine);
            }

      adjustment += -rightScaling.Sum(flank + 1, frame.markers.Length() - 1) * log(RESCALE_AMOUNT);
      }
   }

double JointModel::Evaluate(Vector & point)
   {
   JointLikelihood * engine = qtlMode ? &qtl : &lk;

   double llk = 0;

   static int pass = 0;

   SetModelParameters(point);

   // If we no marker data, computation can be carried out in
   // one pass
   bool haveMarker = currentMarker != -1;

   // First, calculate the likelihood using the
   // full model which allows for trait marker LD
   for (int k = 0; k < families; k++)
      if (!bitsFilter[k])
         {
// #ifdef _DEBUG_
         if (!quiet)
            {
            printf("Pass %d - %s    \r",
                   pass += k == 0, (const char *) mantra[k].family->famid);
            fflush(stdout);
            }
// #endif

         if (!BUFFERING || lookup[k] == k)
            {
            engine->FuzzyScoreVectors(mantra[k]);

            if (engine->MeanProduct(linkage[k]) == 0.0) {
               engine->FuzzyScoreVectors(mantra[k]);
            }

            llk += log(engine->MeanProduct(linkage[k]));
            if (BUFFERING) buffer[k].Copy(*engine);
            if (!haveMarker) llk -= log(engine->WeightedSum());
            }
         else
            {
            llk += log(buffer[lookup[k]].MeanProduct(linkage[k]));
            if (!haveMarker) llk -= log(buffer[lookup[k]].WeightedSum());
            }
         }

   if (!haveMarker)
      return -llk;

   // Next, divide this by the likelihood without taking
   // marker data into consideration (i.e. so that we calculate
   // the conditional probability of the marker data, rather
   // than the joint probability of trait and marker data)
   engine->SelectMarker(-1);
   engine->frequencies[0] = disease[0];
   engine->frequencies[1] = disease[1];
   for (int k = 0; k < families; k++)
      if (!bitsFilter[k])
         {
         if (!BUFFERING || lookup[k] == k)
            {
            engine->FuzzyScoreVectors(mantra[k]);
            llk -= log(engine->WeightedSum());
            if (BUFFERING) buffer[k].Copy(*engine);
            }
         else
            llk -= log(buffer[lookup[k]].WeightedSum());
         }

   return -llk;
   }

void JointModel::SetModelParameters(Vector & point)
   {
   JointLikelihood * engine = qtlMode ? &qtl : &lk;

   engine->SelectMarker(currentMarker);

   disease[0] = disease[1] = 0.0;

   // Retrieve disease-marker haplotype frequencies
   if (engine->frequencies.Length() == 2)
      {
      disease[0] = engine->frequencies[0] = point[2];
      disease[1] = engine->frequencies[1] = 1.0 - point[2];
      }
   else
      {
      int    haplos = engine->frequencies.Length();
      double cumulative = 0.0;

      for (int i = 0; i < haplos / 2 - 1; i++)
         {
         double marker = point[2 + i] * (1.0 - cumulative);
         double conditional = currentModel == 1 ?
                  point.Last() : point[1 + i + engine->frequencies.Length() / 2];

         disease[0] += engine->frequencies[i * 2] = marker * conditional;
         disease[1] += engine->frequencies[i * 2 + 1] = marker * (1.0 - conditional);

         cumulative += marker;
         }

      double marker = 1.0 - cumulative;
      double conditional = point.Last();

      disease[0] += engine->frequencies[haplos - 2] = marker * conditional;
      disease[1] += engine->frequencies[haplos - 1] = marker * (1.0 - conditional);
      }

   if (currentModel == 0)
      {
      if (qtlMode)
         qtl.means.Set(0.0), qtl.variance = 1.0;
      else
         lk.penetrance.Set(PREVALENCE);
      }
   else
      {
      Vector frequencies(3);

      frequencies[0] = disease[0] * disease[0];
      frequencies[1] = 2 * disease[1] * disease[0];
      frequencies[2] = disease[1] * disease[1];

      if (qtlMode)
         {
         qtl.means[0] = point[0];
         qtl.means[1] = point[1];

         ScaleMeans(frequencies, qtl.means, qtl.variance);
         //  ScaleMeans(disease[0], point[0], point[1], qtl.means, qtl.variance);
         }
      else
         {
         lk.penetrance[0] = point[0];
         lk.penetrance[1] = point[1];

         if (FREE)
            ScalePenetrances(PREVALENCE, frequencies, lk.penetrance);
         else if (MULTIPLICATIVE)
            {
            if (disease[0] < 1e-8) disease[0] = 1e-8;

            double minr = max((-(1 - disease[0]) + sqrt(PREVALENCE))/disease[0], 1e-2);

            double minc = max((-(1 - disease[1]) + sqrt(PREVALENCE))/disease[1], 1e-2);
            double maxr = 1.0 / minc;

            double risk = minr + (maxr - minr) * lk.penetrance[0];

            double base = PREVALENCE /
                   (frequencies[0] * risk * risk + frequencies[1] * risk + frequencies[2]);

            lk.penetrance[0] = base * risk * risk;
            lk.penetrance[1] = base * risk;
            lk.penetrance[2] = base;
            }
          else /* ADDITIVE, DOMINANT, RECESSIVE */
            {
            Vector pooled(2);

            if (DOMINANT)
               pooled[0] = frequencies[0] + frequencies[1],
               pooled[1] = frequencies[2];
            else if (RECESSIVE)
               pooled[0] = frequencies[0],
               pooled[1] = frequencies[1] + frequencies[2];
            else if (ADDITIVE)
               pooled[0] = frequencies[0] + frequencies[1] * 0.5,
               pooled[1] = frequencies[2] + frequencies[1] * 0.5;

            // We always have the first allele as the risk increasing allele
            ScalePenetrances(PREVALENCE, pooled, lk.penetrance, currentModel < 3);

            lk.penetrance[2] = lk.penetrance[1];

            if (DOMINANT)
               lk.penetrance[1] = lk.penetrance[0];

            if (ADDITIVE)
               lk.penetrance[1] = (lk.penetrance[2] + lk.penetrance[0]) * 0.5;
            }
         }
      }
   }

void JointModel::SelectModel(int model)
   {
   currentAllele = model < 3 ? -1 : model - 3;
   currentModel  = model > 2 ?  3 : model;

   if (families) BuildLookup(*mantra[0].pedigree);
   }

void JointModel::BuildBitsFilter(Pedigree & ped)
   {
   int familiesSkipped = false;
   int n = ped.familyCount;

   bitsFilter.Dimension(n);
   bitsFilter.Zero();

   for (int i = 0; i < n; i++)
      if (mantra[i].bit_count > MAXBITS)
         {
         bitsFilter[i] = 1;

         if (!familiesSkipped)
            printf("These complex families with >%d bits will be skipped:\n",
                   MAXBITS, familiesSkipped = true);

         printf("   %s [%d bits]\n", (const char *) ped.families[i]->famid,
                                     mantra[i].bit_count);
         }

   if (familiesSkipped)
      printf("\n");
   }

void JointModel::MendelFilter(Pedigree & ped)
   {
   int n = ped.familyCount;

   if (currentMarker == -1)
      return;

   FuzzyInheritanceTree houdini;

   for (int i = 0; i < n; i++)
      if (!bitsFilter[i])
         {
         mantra[i].SelectMarker(currentMarker);
         houdini.FuzzyScoreVectors(mantra[i]);

         if (houdini.FindNonZero() != true)
            {
            mendelFlag = true;

            Family * f = mantra[i].family;
            for (int i = f->first; i <= f->last; i++)
               ped[i].markers[currentMarker].Wipe();
            }
         }
   }

void JointModel::ScalePenetrances(double prevalence,
                                  Vector & frequencies,
                                  Vector & penetrances,
                                  bool first_genotype_increases_risk)
   {
   double complement = 1.0;

   for (int i = 0; i < frequencies.Length() - 1; i++)
      {
      complement -= frequencies[i];

      double lo = max(first_genotype_increases_risk && i == 0 ? prevalence : 0.0,
                      (prevalence - complement) / frequencies[i]);
      double hi = min(1.0, prevalence / frequencies[i]);

      penetrances[i] = lo + penetrances[i] * (hi - lo);
      prevalence     = max(0.0, prevalence - frequencies[i] * penetrances[i]);
      }

   penetrances[frequencies.Length() - 1] = prevalence / complement;
   }

void JointModel::ScaleMeans(Vector & frequencies,  Vector & means, double & var)
   {
   double complement = 1.0;
   double mean = 0.0;

   var = 0.95;

   for (int i = 0; i < frequencies.Length() - 1; i++)
      {
      complement -= frequencies[i];

      double max = sqrt(var / (frequencies[i] + square(frequencies[i]) / complement));

      double delta = means[i] * max;
      double complement_delta = delta * frequencies[i] / complement;

      means[i] = delta + mean;

      mean = mean - complement_delta;
      var -= square(delta) * frequencies[i] + square(complement_delta) * complement;
      }

   means.Last() = mean;
   var += 0.05;
   }

void JointModel::ScaleMeans(double p, double add, double dom,
                            Vector & means, double & residualVariance)
   {
   double q = 1.0 - p;
   double two_pq = 2.0 * p * q;

   double a = add * sqrt( 1.0 / two_pq );
   double d = dom * sqrt( (1.0 - two_pq * a * a) / ((1.0 - two_pq) * two_pq) );

   means[0] = -a;
   means[1] = 0;
   means[2] = a;

   double grandMean = p * p * means[0] + two_pq * means[1] + q * q * means[2];

   means.Add(-grandMean);

   residualVariance = 1.0 - p * p * means[0] * means[0] -
                            two_pq * means[1] * means[1] -
                            q * q * means[2] * means[2];

   // printf( "   p = %.3f,   a = %.3f,   d = %.3f\n"
   //        " g11 = %.3f, g12 = %.3f, g22 = %.3f\n",
   //         p, a, d, means[0], means[1], means[2]);

   if (residualVariance <= 0.0001)
      error("Effect size too large\n"
            "  p = %.3f, a = %.3f, d = %.3f\n",
            p, a, d);
   }


void JointModel::ChooseStartingPoint(Vector & point)
   {
   GetStartingPoint(point);

   Vector best_point = point;
   double best_llk = Evaluate(point);

   for (int i = 0; i < 10; i++)
      {
      GetStartingPoint(point, i);

      double llk = Evaluate(point);

      if (llk < best_llk)
         {
         best_point = point;
         best_llk = llk;
         }
      }

   point = best_point;
   }

void JointModel::PrintLOD(double likelihoodRatio, int df)
   {
   double chisq  = 2.0 * likelihoodRatio;

   if (chisq < 0.0 && chisq > -1e-6) chisq = 0.0;

   double lod    = chisq / (2.0 * log(10.0));
   double pvalue = chisq < 0 ? 1.0 : chidist(chisq, df);

   printf("%6.2f %2d ", lod, df);

   if (chisq < 0.0)
      printf("%6s ", "*");
   else if (pvalue > 1.5e-4)
      printf("%6.*f ", int(log(pvalue*0.06666666)*-0.4343), pvalue);
   else
      printf("%6.0e ", pvalue);

   fflush(stdout);
   }

void JointModel::BuildLookup(Pedigree & ped)
   {
   if (!BUFFERING)
      return;

   StringIntHash  lookupIds;
   String         lookupString;

   int n = ped.familyCount;

   lookup.Dimension(n);
   if (qtlMode)
      {
      lookup.SetSequence(0,1);
      return;
      }

   for (int i = 0; i < n; i++)
      if (!bitsFilter[i])
         {
         Family * f = ped.families[i];

         // Clear the lookup string
         lookupString = "";

         // The first portion of each string indicates fathers for each individual
         for (int j = f->founders; j < f->count; j++)
            {
            lookupString += ped[f->path[j]].father->traverse;
            lookupString += ".";
            }

         // The next portion of each string indicates mothers for each individual
         for (int j = f->founders; j < f->count; j++)
            {
            lookupString += ped[f->path[j]].mother->traverse;
            lookupString += ".";
            }

         // If we are analysing a marker location, then list genotypes
         if (currentMarker >= 0)
            for (int j = 0; j < f->count; j++)
               {
               lookupString += ped[f->path[j]].markers[currentMarker].BinaryCoded();
               lookupString += ".";
               }

         // Finally, if we are analyzing a trait locus ...
         if (currentModel > 0)
            for (int j = 0; j < f->count; j++)
               {
               lookupString += ped[f->path[j]].affections[currentAffection];
               lookupString += ".";
               }

         // Check if we have come across a family with similar structure
         int id = lookupIds.Integer(lookupString);

         // Assign an id for each unique family configuration
         if (id >= 0)
            lookup[i] = id;
         else
            {
            lookup[i] = i;
            lookupIds.SetInteger(lookupString, i);
            }
         }
   }
   
 
